
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "EnzoGridHierarchy.hh"
#include <AMRwriter.hh>

#define STAR_PARTICLE_NUMBER_START 1000000000
// #define STAR_PARTICLE_NUMBER_START 10000000
#define STAR_PARTICLE_INDEX STAR_PARTICLE_NUMBER_START

EnzoGridHierarchy::~EnzoGridHierarchy(){
  for(int i=0;i<StarDataNames.getSize();i++) 
    if(StarDataNames[i])
      delete StarDataNames[i];
  // should probably delete all of the grids too...
}

int EnzoGridHierarchy::readHierarchy(char *file){
  EnzoGridInfo *g;
  FILE *hierarchyfile;
  hierarchyfile = fopen(file,"r");
  if(!hierarchyfile){
    printf("could not open %s for reading\n",file);
    return 0;
  }
  //puts("setup grids");
  do {
    g = new EnzoGridInfo(hierarchyfile,grids.getData()); // no grid stuff here
    //puts("next");
    if(g->isValid()) {
      //puts("valid");
      grids.append(g);
      // And do some other g stuff (like nextgridthislevel)
      
    }
  } while(g->isValid());
  delete g; // last one is not valid
  // fix up the grid cross-references
  printf("Ngrids=%u\n",grids.getSize());
  for(int i=0;i<grids.getSize();i++){
    g = grids[i];
    if(g->nextGridNextLevel > 0) {
      g->nextlev = grids[g->nextGridNextLevel-1];
      g->nextlev->prevlev = g;
    }
    else 
      g->nextlev=0;
    if(g->nextGridThisLevel > 0){
      g->next = grids[g->nextGridThisLevel-1];
      g->next->prev = g;
    }
    else 
      g->next=0;
  }
  //printf("now assignLevel\n");
  if(grids.getSize()>0 && grids[0])
    grids[0]->assignLevel(0); // recursively assign level info
  fclose(hierarchyfile);
  if(!grids.getSize() || !grids[0]){
    printf("where the hell is the hierarchy?\n");
    return 0;
  }
  // nuthin else to read.  Now assign the levels in the hierarchy
  // find max level first
  {
    int maxlevel=0;
    for(int i=0;i<grids.getSize();i++){
      if(maxlevel<grids[i]->level) maxlevel=grids[i]->level;
    }
    levels.setSize(maxlevel+1);
  }
  // index grids by level
  {
    for(int i=0;i<grids.getSize();i++)
      if(grids[i]->level>=0 && grids[i]->level<levels.getSize())
	(levels[grids[i]->level]).append(grids[i]);
      else
	printf("Bogus level %d\n",grids[i]->level);
  }
  // now create parent-child relationships between the grids.
  // needs a "contains" method (like in IEEEIO:AMRreader)
  {
    for(int l=1;l<levels.getSize();l++){
      for(int i=0;i<(levels[l-1]).getSize();i++){
	EnzoGridInfo *parent=(levels[l-1])[i];
	for(int j=0;j<(levels[l]).getSize();j++){
	  EnzoGridInfo *child=(levels[l])[j];
	  // find all of my children and set parent relations
	  if(parent->contains(child)) {
	    parent->children.append(child);
	    child->parent=parent;
	  }
	}
      }
    }
  } 
  char *starfile = new char[strlen(file)+1];
  strcpy(starfile,file);
  strcpy(strrchr(starfile,'.'),".stars");
  readStarData(starfile,
	       StarDataNames,
	       StarData);
  return 1;
}

// void EnzoGridHierarchy::outputXML(FILE *file){
// }
int EnzoGridHierarchy::outputInfo(FILE *file){
  // Lets scan for file contents (find first available file
  {
    for(int i=0;i<grids.getSize();i++){
      FlexArray<char *> names;
      if(grids[i]->readNames(names)){
	int hasparticles=0;
	printf("The following Field Quantities are present\n");
	for(int j=0;j<names.getSize();j++){
	  if(!strstr(names[j],"particle"))
	    printf("\t%s\n",names[j]);
	  else
	    hasparticles=1;
	}
	if(hasparticles){
	  printf("The following Particle Quantities are present\n");
	  {for(int j=0;j<names.getSize();j++){
	    if(!strstr(names[j],"position") && strstr(names[j],"particle")){
	      if(strstr(names[j],"velocity")){
		if(names[j][18]=='x')
		  printf("\tvelocity_{x,y,z}\n");
	      }
	      else {
		char *s=names[j]+9;
		printf("\t%s\n",s);
	      }
	    }
	  }
	  }}
	{for(int j=0;j<names.getSize();j++)
	  delete names[j];
	}
	break;
      }
      // else puts("failed: try again");
    }
  }
  // Number of grids per level
  printf("%u levels\n",levels.getSize());
  for(int i=0;i<levels.getSize();i++)
    printf("\tLevel[%u] has %u grids\n",i,(levels[i]).getSize());
  return 1;
}

// need to do this one-per-data field
// need a file array
int EnzoGridHierarchy::writeAllJad(char *fileprefix, int step){
  // open JAD file
  double dx[3];
  // IEEEIO *file = 0;
  FlexArray<char *> names; // get all field names
  FlexArray<IEEEIO*> files;
  FlexArray<AMRwriter*> amr;
  EnzoGridInfo *g=(levels[0])[0];
  g->readNames(names);
  printf("Toplevel delta=%lf\n",g->dx);
  dx[0]=dx[1]=dx[2]=g->dx;
  files.setSize(names.getSize());
  amr.setSize(names.getSize());
  {
    char *filename = new char[strlen(fileprefix)+64];
    for(int i=0;i<names.getSize();i++){
      sprintf(filename,"%s.%s.raw",fileprefix,names[i]);
      if(step>0)
	files[i] = new IEEEIO(filename,IObase::Append); // open file for append
      else
	files[i] = new IEEEIO(filename,IObase::Write); // or truncate
      amr[i] = new AMRwriter(*(files[i]));
      amr[i]->setTopLevelParameters(3,g->minext,dx,1.0,levels.getSize());
      amr[i]->setRefinement(1, // timerefinement (was 2 (normally is 2)
			    2, // spatialrefinement
			    1);// gridplacementrefinement 
      // (w/respect to spatial refinement... 
      // 1 means it is the same as the spatial refinement)
      // really should be 2 though since the grid is staggered
      amr[i]->setTime(step); // assume snapshot of hierarchy at timestep 0
    }
    delete filename;
  }
  {
    FlexArray<float> data;
    for(int l=0;l<levels.getSize();l++){
      // need a non-allocating readGrid
      printf("\nWriting level %u\n",l);
      for(int j=0;j<amr.getSize();j++) 
	amr[j]->setLevel(l);
      for(int i=0;i<(levels[l]).getSize();i++){
	g = (levels[l])[i];
	fprintf(stderr,"*");
	// float *data = g->readGrid(fieldname);
	for(int n=0;n<names.getSize();n++){
	  g->readGrid(names[n],data);
	  fprintf(stderr,".");
	  amr[n]->write(g->minext,g->dims,data.getData());
	}
	//}
      }
    }
  }
  puts("done");
  {for(int i=0;i<names.getSize();i++) delete names[i];}
  {for(int i=0;i<amr.getSize()  ;i++) delete amr[i];  }
  {for(int i=0;i<files.getSize();i++) delete files[i];}
  return 1;
}

// need to do this one-per-data field
// need a file array
int EnzoGridHierarchy::writeJad(char *fileprefix,char *fieldname,int step){
  // open JAD file
  char *filename = new char[strlen(fileprefix)+64];
  sprintf(filename,"%s.%s.amr.raw",fileprefix,fieldname);
  double dx[3];
  IEEEIO *file =0;
  if(step>0)
    file = new IEEEIO(filename,IObase::Append); // open file for append or trunc
  else
    file = new IEEEIO(filename,IObase::Write);
  delete filename;
  AMRwriter *amr = new AMRwriter(*file);
  // AMRwriter amrfile(outfile);
  EnzoGridInfo *g=(levels[0])[0];
  printf("Toplevel delta=%lf\n",g->dx);
  // amrfile.setRank(rank);
  dx[0]=dx[1]=dx[2]=g->dx;
  amr->setTopLevelParameters(3,g->minext,dx,1.0,levels.getSize());
  // int refinement;
  // int RefineBy=2; // for now...
  amr->setRefinement(1, // timerefinement (was 2 (normally is 2)
		     2, // spatialrefinement
		     1); // gridplacementrefinement 
  // (w/respect to spatical refinement... 
  // 1 means it is the same as the spatial refinement)
  // really should be 2 though since the grid is staggered
  amr->setTime(step); // assume snapshot of hierarchy at timestep 0
  {
    FlexArray<float> data;
    for(int l=0;l<levels.getSize();l++){
      // need a non-allocating readGrid
      amr->setLevel(l);
      for(int i=0;i<(levels[l]).getSize();i++){
	g = (levels[l])[i];
	// float *data = g->readGrid(fieldname);
	g->readGrid(fieldname,data);
	amr->write(g->minext,g->dims,data.getData());
	// amr->write(g->minext,g->dims,data);
	// delete data;
      }
    }
  }
  delete amr;
  delete file;
  return 1;
}

// need a downsampling mode of operation
int EnzoGridHierarchy::writeGasParticles(char *filename,char *fieldname,int step){
  //  double dx[3];
  IObase *file =0;
  printf("filename=%s\n",filename);
  if(step>0)
    file = new IEEEIO(filename,IObase::Append); // open file for append or trunc
  else
    file = new IEEEIO(filename,IObase::Write);
  EnzoGridInfo *g=(levels[0])[0];
  printf("Toplevel delta=%lf\n",g->dx);
  FlexArray<float> data,coordsout,dataout;
  for(int l=0;l<levels.getSize();l++){ // for each level
    // lets check each point against children grids to see if we should write it
    // allow particle skipping based on dims of child grid eventually
    for(int n=0;n<(levels[l]).getSize();n++){
      int *dims;
      float coords[3];
      double range[3],fdims[3],*origin;
      g = (levels[l])[n];
      fprintf(stderr,"reading l:%2u g:%03u\r",l,n);
      g->readGrid(fieldname,data);
      dims = g->dims;
      origin = g->minext;
      // fprintf(stderr,"copying \r");
      for(int i=0;i<3;i++) {
	range[i] = g->maxext[i] - g->minext[i];
	fdims[i] = (float)((dims[i]>1)?dims[i]-1:1);
      }
      if(data.getSize()!=dims[0]*dims[1]*dims[2]) {
	printf("No Data!!! dims=%u:%u:%u datasize=%u\n",dims[0],dims[1],dims[2],data.getSize());
	continue;
      }
      for(int i=0,iz=0;iz<g->dims[2];iz++){
	coords[2] = origin[2] + range[2]*(float)(iz)/fdims[2];
	for(int iy=0;iy<g->dims[1];iy++){
	  coords[1] =  origin[1] + range[1]*(float)(iy)/fdims[1];
	  for(int ix=0;ix<g->dims[0];ix++,i++){
	    coords[0] =  origin[0] + range[0]*(float)(ix)/fdims[0];
	    int skip = 0;
	    for(int child=0;child<g->children.getSize();child++){
	      if((g->children[child])->contains(coords[0],coords[1],coords[2])){
		// skip this point and a bunch more
		// for now we'll ignore (each point gets compared to all children!)
		skip=1;
	      }
	    }
	    if(skip) continue;
	    coordsout.append(coords,3);
	    dataout.append(data[i]);
	  }
	}
      }
    }
  }
  int ldims[2];
  ldims[0]=3;
  ldims[1]=dataout.getSize();
  printf("\nTotalsize = %u particles\nWriting\n",dataout.getSize());
  file->write(IObase::Float32,2,ldims,coordsout.getData());
  file->writeAttribute(fieldname,IObase::Float32,dataout.getSize(),dataout.getData());
  delete file;
  return 1;
}
// need a downsampling mode of operation
int EnzoGridHierarchy::writeAllGasParticles(char *filename,int step){
  //  double dx[3];
  // char *fieldname = "Density"; // kludge
  IObase *file = 0;
  printf("filename=%s\n",filename);
  if(step>0)
    file = new IEEEIO(filename,IObase::Append); // open file for append or trunc
  else
    file = new IEEEIO(filename,IObase::Write);
  EnzoGridInfo *g=(levels[0])[0];
  printf("Toplevel delta=%lf\n",g->dx);
  typedef FlexArray<float> FloatFlexArray;
  //typedef char StaticString[32];
  FlexArray<FloatFlexArray> fielddata,fielddataout;
  FlexArray<char *> fieldnames;
  FlexArray<float> coordsout; // a number of things
  
  g->readFieldNames(fieldnames); // read all field names
  {
    char *s=new char[strlen("level")+1];
    strcpy(s,"level");
    fieldnames.append(s);
  }
  fielddata.setSize(fieldnames.getSize()); // resize data holders to field size
  fielddataout.setSize(fieldnames.getSize()); // resize data holders to field size
  printf("%u field names\n\n",fieldnames.getSize());

  for(int l=0;l<levels.getSize();l++){ // for each level
    // lets check each point against children grids to see if we should write it
    // allow particle skipping based on dims of child grid eventually
    for(int n=0;n<(levels[l]).getSize();n++){
      int *dims;
      float coords[3];
      double range[3],fdims[3],*origin;
      g = (levels[l])[n];
      fprintf(stderr,"reading l:%2u g:%03u\r",l,n);
      for(int f=0;f<fieldnames.getSize()-1;f++)
	g->readGrid(fieldnames[f],fielddata[f]);
      for(int i=0,last=IObase::nElements(3,g->dims),
	    fnum=fieldnames.getSize()-1;i<last;i++)
	(fielddata[fnum]).append(g->level); // store level information
      dims = g->dims;
      origin = g->minext;
      // fprintf(stderr,"copying \r");
      for(int i=0;i<3;i++) {
	range[i] = g->maxext[i] - g->minext[i];
	fdims[i] = (float)((dims[i]>1)?dims[i]-1:1);
      }
      if(fielddata[0].getSize()!=dims[0]*dims[1]*dims[2]) {
	printf("No Data!!! dims=%u:%u:%u datasize=%u\n",dims[0],dims[1],dims[2],fielddata[0].getSize());
	continue;
      }
      for(int i=0,iz=0;iz<g->dims[2];iz++){
	coords[2] = origin[2] + range[2]*(float)(iz)/fdims[2];
	for(int iy=0;iy<g->dims[1];iy++){
	  coords[1] =  origin[1] + range[1]*(float)(iy)/fdims[1];
	  for(int ix=0;ix<g->dims[0];ix++,i++){
	    coords[0] =  origin[0] + range[0]*(float)(ix)/fdims[0];
	    int skip = 0;
	    for(int child=0;child<g->children.getSize();child++){
	      if((g->children[child])->contains(coords[0],coords[1],coords[2])){
		// skip this point and a bunch more
		// for now we'll ignore (each point gets compared to all children!)
		skip=1;
	      }
	    }
	    if(skip) continue;
	    coordsout.append(coords,3);
	    for(int f=0;f<fielddataout.getSize();f++)
	      fielddataout[f].append((fielddata[f])[i]);
	  }
	}
      }
    }
  }
  int ldims[2];
  ldims[0]=3;
  ldims[1]=fielddataout[0].getSize();
  printf("\nTotalsize = %u particles\nWriting\n",fielddataout[0].getSize());
  file->write(IObase::Float32,2,ldims,coordsout.getData());
  for(int f=0;f<fieldnames.getSize();f++){
    file->writeAttribute(fieldnames[f],
			 IObase::Float32,fielddataout[f].getSize(),
			 fielddataout[f].getData());
    delete fieldnames[f]; // delete all of the names to clean up
  }
  delete file;
  return 1;
}

int EnzoGridHierarchy::writeDMParticles(char *filename,char *fieldname,int step){
  return 0;
}
int EnzoGridHierarchy::writeAllDMParticles(char *filename,int step){
  puts("inside writeAllDM");
  // get all particle names
  // read in all data from all grids into tmp arrays
  // only append data that has indexes less than spec
  // (also compute mass from density)
  int iindex,xindex,yindex,zindex,dindex;
  FlexArray<float> data,xc,yc,zc,leveldata,density; // temporary arrays
  FlexArray<int> index,outindex;
  FlexArray<float> coords;
  typedef FlexArray<float> FloatFlexArray;
  FlexArray<FloatFlexArray> dataout;
  FlexArray<char *> names;
  IObase *file = 0;
  printf("filename=%s\n",filename);
  if(step>0)
    file = new IEEEIO(filename,IObase::Append); // open file for append or trun
  else
    file = new IEEEIO(filename,IObase::Write);
  EnzoGridInfo *grid=(levels[0])[0];
  grid->readParticleNames(names); // read all field names
  dataout.setSize(names.getSize()); // resize data holders to field size
  printf("Particle field names are\n");
  for(int i=0;i<names.getSize();i++) {
    printf("\t%s\n",names[i]);
    if(strstr(names[i],"index")) iindex=i;
    else if(strstr(names[i],"position")){
      if(strchr(names[i],'x')) xindex=i;
      else if(strchr(names[i],'y')) yindex=i;
      else if(strchr(names[i],'z')) zindex=i;
    }
    else if(strstr(names[i],"mass")) dindex=i;
  }
  // printf("indices are i=%u x=%u y=%u z=%u\n",iindex,xindex,yindex,zindex);
  // cycle through each level
  for(int l=0;l<levels.getSize();l++){ // for each level
    for(int g=0;g<(levels[l]).getSize();g++){ // for each grid at that level
      grid = (levels[l])[g];
      // grid = (levels[0])[0];
      if(!grid) fprintf(stderr,"no grid!\n");
      grid->readQuantity(names[iindex],index);
      // copy into an integer array? (outindex?) or float index?
      // for now, lets ignore it?
      for(int i=0;i<index.getSize();i++)
	if(index[i]<STAR_PARTICLE_INDEX) {
	  dataout[iindex].append((float)(index[i]));
	  outindex.append(index[i]);
	}
      grid->readQuantity(names[xindex],xc); // xcoords
      grid->readQuantity(names[yindex],yc); // ycoords
      grid->readQuantity(names[zindex],zc); // zcoords
      for(int i=0;i<index.getSize();i++){
	if(index[i]<STAR_PARTICLE_INDEX){
	  coords.append(xc[i]);
	  coords.append(yc[i]);
	  coords.append(zc[i]);
	  leveldata.append(grid->level);
	}
      }
      fprintf(stderr,"\rL:%2u G:%03u %04u particles",l,g,data.getSize());
      for(int n=0;n<names.getSize();n++){
	// now we start cycling through the dataset to read each name using
	// readQuantity
	if(strstr(names[n],"index") || strstr(names[n],"position")) 
	  continue;
	grid->readQuantity(names[n],data);
	// now copy based on its index
	//(if it is a coordinate, we need to append to a single place)
	// but it has to be accumulated first
	//printf("indexsize=%u datasize=%u\n",index.getSize(),data.getSize());
	for(int i=0;i<index.getSize() && i<data.getSize();i++){
	  if(index[i]<STAR_PARTICLE_INDEX) 
	    dataout[n].append(data[i]);
	}
      }
      fprintf(stderr,"\tTotal DM=%u:%u",
	      dataout[dindex].getSize(),coords.getSize()/3);
    }
  }
  density.append(dataout[dindex].getData(),dataout[dindex].getSize());
  
  
  for(int i=0;i<dataout[dindex].getSize();i++)
    (dataout[dindex])[i] = density[i] * 2.763e+11 * 2.0/(pow(2.0,3.0*leveldata[i]));
  {
    float mrange[2];
    mrange[0] = mrange[1]= (dataout[dindex])[0];
    for(int i=0;i<dataout[dindex].getSize();i++){
      float val = (dataout[dindex])[i];
      if(val<mrange[0]) mrange[0]=val;
      if(val>mrange[1]) mrange[1]=val;
    }
    printf("\nMass Range is %g:%g\n",mrange[0],mrange[1]);
  }  
  {
    float mrange[2];
    mrange[0] = mrange[1]= density[0];
    for(int i=0;i<density.getSize();i++){
      float val = density[i];
      if(val<mrange[0]) mrange[0]=val;
      if(val>mrange[1]) mrange[1]=val;
    }
    printf("\nDensity Range is %g:%g\n",mrange[0],mrange[1]);
  }
  // Now that we have big-honking arrays, its time to dump them to disk
  puts("\nDone\n");
  int ldims[2];
  ldims[0]=3;
  ldims[1]=coords.getSize()/3;
  printf("\nDM Totalsize = %u particles\nWriting\n",coords.getSize()/3);
  file->write(IObase::Float32,2,ldims,coords.getData());
  for(int f=0;f<names.getSize();f++){
    if(!strstr(names[f],"position") && f!=iindex)
      file->writeAttribute(names[f],
			   IObase::Float32,dataout[f].getSize(),
			   dataout[f].getData());
    delete names[f]; // delete all of the names to clean up
  }
  file->writeAttribute("level",IObase::Float32,
		       leveldata.getSize(),leveldata.getData());
  file->writeAttribute("density",IObase::Float32,
		       density.getSize(),density.getData());
  file->writeAttribute("findex",IObase::Float32,
		       dataout[iindex].getSize(),dataout[iindex].getData());
  file->writeAttribute("index",IObase::Int32,
		       outindex.getSize(),outindex.getData());
  // should provide an integer index in output as well eventually
  delete file;
  return 1;
}
int EnzoGridHierarchy::writeStarParticles(char *filename,char *fieldname,int step){
  return 0;
}
int EnzoGridHierarchy::writeAllStarParticles(char *filename,int step){  
  // get all particle names
  // read in all data from all grids into tmp arrays
  // only append data that has indexes less than spec
  // (also compute mass from density)
  int iindex,xindex,yindex,zindex,dindex;
  FlexArray<float> data,xc,yc,zc,leveldata,density; // temporary arrays
  FlexArray<int> index,outindex;
  FlexArray<float> coords;
  typedef FlexArray<float> FloatFlexArray;
  FlexArray<FloatFlexArray> dataout;
  FlexArray<char *> names;
  IObase *file = 0;
  printf("filename=%s\n",filename);
  if(step>0)
    file = new IEEEIO(filename,IObase::Append); // open file for append or trun
  else
    file = new IEEEIO(filename,IObase::Write);
  EnzoGridInfo *grid=(levels[0])[0];
  grid->readParticleNames(names); // read all field names
  dataout.setSize(names.getSize()); // resize data holders to field size
  for(int i=0;i<names.getSize();i++) {
    if(strstr(names[i],"index")) iindex=i;
    else if(strstr(names[i],"position")){
      if(strchr(names[i],'x')) xindex=i;
      else if(strchr(names[i],'y')) yindex=i;
      else if(strchr(names[i],'z')) zindex=i;
    }
    else if(strstr(names[i],"mass")) dindex=i;
  }
  int numgrids=0,alltotal=0;
  // cycle through each level
  for(int l=0;l<levels.getSize();l++){ // for each level
    for(int g=0;g<(levels[l]).getSize();g++){ // for each grid at that level
      numgrids++;
      grid = (levels[l])[g];
      grid->readQuantity(names[iindex],index);
      // copy into an integer array? (outindex?) or float index?
      // for now, lets ignore it?
      for(int i=0;i<index.getSize();i++){
	if(index[i]>=STAR_PARTICLE_INDEX){
	  dataout[iindex].append((float)(index[i]));
	  outindex.append(index[i]);
	}
      }
      grid->readQuantity(names[xindex],xc); // xcoords
      grid->readQuantity(names[yindex],yc); // ycoords
      grid->readQuantity(names[zindex],zc); // zcoords
      for(int i=0;i<index.getSize();i++){
	if(index[i]>=STAR_PARTICLE_INDEX){
	  coords.append(xc[i]);
	  coords.append(yc[i]); 
	  coords.append(zc[i]);
	  leveldata.append(grid->level);
	}
      }
      
      for(int n=0;n<names.getSize();n++){
	// now we start cycling through the dataset to read each name using
	// readQuantity
	if(strstr(names[n],"index") || strstr(names[n],"position")) 
	  continue;
	grid->readQuantity(names[n],data);
	if(n==dindex){
	  fprintf(stderr,"F:[%s]  L:%2u G:%03u %04u particles",grid->filename,l,g,data.getSize());
	  alltotal+=data.getSize();
	}
	// now copy based on its index
	//(if it is a coordinate, we need to append to a single place)
	// but it has to be accumulated first
	for(int i=0;i<index.getSize() && i<data.getSize();i++){
	  if(index[i]>=STAR_PARTICLE_INDEX) 
	    dataout[n].append(data[i]);
	}
      }
      fprintf(stderr," : Total Stars=%u\r",
	      dataout[dindex].getSize());
    }
  }
  printf("\nNumber of grids=%u total particles=%u\n",numgrids,alltotal);
  density.append(dataout[dindex].getData(),dataout[dindex].getSize());
  for(int i=0;i<dataout[dindex].getSize();i++)
    (dataout[dindex])[i] = density[i] * 2.763e+11 * 2.0/(pow(2.0,3.0*leveldata[i]));
  {
    float mrange[2];
    mrange[0] = mrange[1]= (dataout[dindex])[0];
    for(int i=0;i<dataout[dindex].getSize();i++){
      float val = (dataout[dindex])[i];
      if(val<mrange[0]) mrange[0]=val;
      if(val>mrange[1]) mrange[1]=val;
    }
    printf("\nMass Range is %g:%g\n",mrange[0],mrange[1]);
  } 
  {
    float mrange[2];
    mrange[0] = mrange[1] = density[0];
    for(int i=0;i<density.getSize();i++){
      float val = density[i];
      if(val<mrange[0]) mrange[0]=val;
      if(val>mrange[1]) mrange[1]=val;
    }
    printf("\nDensity Range is %g:%g\n",mrange[0],mrange[1]);
  }

  
  /* Rescale the indices */
  {
    int minindex=outindex[0],maxindex=outindex[0];
    for(int i=0;i<outindex.getSize();i++){
      int val=outindex[i];
      if(minindex>val) minindex=val;
      if(maxindex<val) maxindex=val;
    }
    for(int i=0;i<outindex.getSize();i++){
      outindex[i]-=minindex;
      (dataout[iindex])[i]=(float)(outindex[i]);
    }
  }
  // Now we need to reindex in the Stars Data
  {
    int sz=dataout.getSize();
    dataout.setSize(dataout.getSize() + this->StarData.getSize());
    for(int i=0;i<StarData.getSize();i++,sz++){
      dataout[sz].setSize(outindex.getSize());
      char *s = new char[strlen(StarDataNames[i])+1];
      strcpy(s,StarDataNames[i]);
      names.append(s);
      mapIndexedData(outindex.getSize(),outindex.getData(),this->StarData[i].getData(),dataout[sz].getData());
    }
  }
  
  // Now that we have big-honking arrays, its time to dump them to disk
  puts("now write the stuff out!");
  int ldims[2];
  ldims[0]=3;
  ldims[1]=coords.getSize()/3;
  printf("\nStars Totalsize = %u particles\nWriting\n",coords.getSize()/3);
  file->write(IObase::Float32,2,ldims,coords.getData());
  for(int f=0;f<names.getSize();f++){
    if(!strstr(names[f],"position") && f!=iindex)
      file->writeAttribute(names[f],
			   IObase::Float32,dataout[f].getSize(),
			   dataout[f].getData());
    delete names[f]; // delete all of the names to clean up
  }
  file->writeAttribute("level",IObase::Float32,
		       leveldata.getSize(),leveldata.getData());
  file->writeAttribute("density",IObase::Float32,
		       density.getSize(),density.getData());   
  //printf("returnval = %d\n",
  file->writeAttribute("findex",IObase::Float32,
		       dataout[iindex].getSize(),dataout[iindex].getData());
  //);
  // printf("returnval = %d\n",
  file->writeAttribute("index",IObase::Int32,
		       outindex.getSize(),outindex.getData());
	 //);
  // should provide an integer index in output as well eventually
  delete file;
  return 1;
}

// read info from hierarchyfilebase.stars
int EnzoGridHierarchy::readStarData(char *starfile,
				    FlexArray<char *> &names,
				    FlexArray<FlexArrayFloat> &data){
  HDFIO *infile = new HDFIO(starfile,IObase::Read);
  if(!infile->isValid()) {
    delete infile;
    printf("Could not open starfile [%s]\n",starfile);
    return 0;
  }
  names.setSize(0);
  data.setSize(0);
  printf("Stars ndatasets=%u\n",infile->nDatasets());
  for(int nds=0,i=0;i<infile->nDatasets();i++){
    int rank,dims[3];
    char name[128];
    IObase::DataType dt;
    infile->readInfo(name,dt,rank,dims);
    if(infile->isCoord()) continue;
    if(dt!=IObase::Float32) continue;
    if(rank!=1) continue;
    data.setSize(data.getSize()+1); // increase size
    for(int a=0;a<infile->nAttributes();a++){
      char nname[128];
      infile->readAttributeInfo(a,nname,dt,rank);
      if(!strcmp(nname,"long_name")){
	infile->readAttribute(a,name);
	name[rank]='\0'; // null terminate
      }
    }
    if(strlen(name)<=1) continue;
    char *s= new char[strlen(name)+1];
    strcpy(s,name);
    names.append(s);
    (data[nds]).setSize(dims[0]);
    infile->read(data[nds].getData());
    nds++;
  }
  printf("----------\nNumberOfStarComponents=%u\n",data.getSize());
  for(int i=0;i<names.getSize();i++){
    printf("Star Component[%u] name=[%s] ndata=%u\n",i,names[i],data[i].getSize());
  }
  delete infile;
  return 1;
}

// use indices to map a inputdata to mappeddata.
// will work in one shot.
// find min/max of indices and zero them out
int EnzoGridHierarchy::mapIndexedData(int length,
				      int *indices,
				      float *inputdata,
				      float *outputdata){
  int minindex=indices[0],maxindex=indices[0];
  for(int i=0;i<length;i++){
    int val=indices[i];
    if(minindex>val) minindex=val;
    if(maxindex<val) maxindex=val;
  }
  if((maxindex-minindex) > length) {
    printf("Error in mapIndexedData.  Index is out of range.\n");
    printf("\tminindex=%u maxindex=%u range=%u length of array = %u\n",
	   minindex,maxindex,maxindex-minindex,length);
    return 0;
  }
  for(int i=0;i<length;i++)
    outputdata[i] = inputdata[indices[i]-minindex];
  return 1; // success
}




int EnzoGridHierarchy::writeAllDMParticlesMPI(char *filename,int step){
  puts("inside writeAllDM");
  // get all particle names
  // read in all data from all grids into tmp arrays
  // only append data that has creation time values=-99999 (dark matter)
  // (also compute mass from density)
    int iindex,xindex,yindex,zindex,dindex,cindex;
    FlexArray<float> data,xc,yc,zc,leveldata,density; // temporary arrays
    FlexArray<int> outindex,indexval;
    FlexArray<float> crindex,coords;

  typedef FlexArray<float> FloatFlexArray;
  FlexArray<FloatFlexArray> dataout;
  FlexArray<char *> names;
  IObase *file = 0;
  printf("filename=%s\n",filename);
  if(step>0)
    file = new IEEEIO(filename,IObase::Append); // open file for append or trun
  else
    file = new IEEEIO(filename,IObase::Write);
  EnzoGridInfo *grid=(levels[0])[0];
  grid->readParticleNames(names); // read all field names
  dataout.setSize(names.getSize()); // resize data holders to field size
  printf("Particle field names are\n");
  for(int i=0;i<names.getSize();i++) {
    printf("\t%s\n",names[i]);
    if(strstr(names[i],"index")) iindex=i;
    else if(strstr(names[i],"position")){
      if(strchr(names[i],'x')) xindex=i;
      else if(strchr(names[i],'y')) yindex=i;
      else if(strchr(names[i],'z')) zindex=i;
    }
    else if(strstr(names[i],"creation")) cindex=i;
    else if(strstr(names[i],"mass")) dindex=i;
  }
  //printf("dindex = %d \n",dindex);
  //printf("cindex = %d \n",cindex);

  // printf("indices are i=%u x=%u y=%u z=%u\n",iindex,xindex,yindex,zindex);
  // cycle through each level
  for(int l=0;l<levels.getSize();l++){ // for each level
    for(int g=0;g<(levels[l]).getSize();g++){ // for each grid at that level
      grid = (levels[l])[g];
      // grid = (levels[0])[0];
      if(!grid) fprintf(stderr,"no grid!\n");
      grid->readQuantity(names[iindex],indexval);
      grid->readQuantity(names[cindex],crindex);
      // copy into an integer array? (outindex?) or float index?
      // for now, lets ignore it?
      for(int i=0;i<crindex.getSize();i++)
	if(crindex[i]<0) {

	  dataout[iindex].append((float)(indexval[i]));
	  outindex.append(indexval[i]);
	}
      grid->readQuantity(names[xindex],xc); // xcoords
      grid->readQuantity(names[yindex],yc); // ycoords
      grid->readQuantity(names[zindex],zc); // zcoords
      for(int i=0;i<crindex.getSize();i++){
	if(crindex[i]<0){
	  coords.append(xc[i]);
	  coords.append(yc[i]);
	  coords.append(zc[i]);
	  leveldata.append(grid->level);
	}
      }
      fprintf(stderr,"\rL:%2u G:%03u %04u particles",l,g,data.getSize());
      for(int n=0;n<names.getSize();n++){
	// now we start cycling through the dataset to read each name using
	// readQuantity
	if(strstr(names[n],"index") || strstr(names[n],"position")) 
	  continue;
	grid->readQuantity(names[n],data);
	// now copy based on its index
	//(if it is a coordinate, we need to append to a single place)
	// but it has to be accumulated first
	//printf("indexsize=%u datasize=%u\n",index.getSize(),data.getSize());
	for(int i=0;i<crindex.getSize() && i<data.getSize();i++){
	  if(crindex[i]<0) 
	    dataout[n].append(data[i]);
	}
      }
      fprintf(stderr,"\tTotal DM=%u:%u",
	      dataout[dindex].getSize(),coords.getSize()/3);
    }
  }
  density.append(dataout[dindex].getData(),dataout[dindex].getSize());
  
  for(int i=0;i<dataout[dindex].getSize();i++)
    (dataout[dindex])[i] = density[i] * 2.763e+11 * 2.0/(pow(2.0,3.0*leveldata[i]));
  {
    float mrange[2];
    mrange[0] = mrange[1]= (dataout[dindex])[0];
    for(int i=0;i<dataout[dindex].getSize();i++){
      float val = (dataout[dindex])[i];
      if(val<mrange[0]) mrange[0]=val;
      if(val>mrange[1]) mrange[1]=val;
    }
    printf("\nMass Range is %g:%g\n",mrange[0],mrange[1]);
  }  
  {
    float mrange[2];
    mrange[0] = mrange[1]= density[0];
    for(int i=0;i<density.getSize();i++){
      float val = density[i];
      if(val<mrange[0]) mrange[0]=val;
      if(val>mrange[1]) mrange[1]=val;
    }
    printf("\nDensity Range is %g:%g\n",mrange[0],mrange[1]);
  }
  // Now that we have big-honking arrays, its time to dump them to disk
  puts("\nDone\n");
  int ldims[2];
  ldims[0]=3;
  ldims[1]=coords.getSize()/3;
  printf("\nDM Totalsize = %u particles\nWriting\n",coords.getSize()/3);
  file->write(IObase::Float32,2,ldims,coords.getData());
  for(int f=0;f<names.getSize();f++){
    if(!strstr(names[f],"position") && !strstr(names[f],"creation_time") && !strstr(names[f],"dynamical_time") && !strstr(names[f],"metallicity_fraction") && f!=iindex)
      file->writeAttribute(names[f],
			   IObase::Float32,dataout[f].getSize(),
			   dataout[f].getData());
    delete names[f]; // delete all of the names to clean up
  }
  file->writeAttribute("level",IObase::Float32,
		       leveldata.getSize(),leveldata.getData());
  file->writeAttribute("density",IObase::Float32,
		       density.getSize(),density.getData());
  file->writeAttribute("findex",IObase::Float32,
		       dataout[iindex].getSize(),dataout[iindex].getData());
  file->writeAttribute("index",IObase::Int32,
		       outindex.getSize(),outindex.getData());
  // should provide an integer index in output as well eventually
  delete file;
  return 1;
}


int EnzoGridHierarchy::writeAllStarParticlesMPI(char *filename,int step){  
  // get all particle names
  // read in all data from all grids into tmp arrays
  // only append data that has creation times > 0 (star particles)
  // (also compute mass from density)
  int iindex,xindex,yindex,zindex,dindex,cindex; 

  FlexArray<float> data,xc,yc,zc,leveldata,density;  // temporary arrays
   FlexArray<int> outindex,indexval;
   FlexArray<float> crindex,coords;

  typedef FlexArray<float> FloatFlexArray;
  FlexArray<FloatFlexArray> dataout;
  FlexArray<char *> names;
  IObase *file = 0;
  printf("filename=%s\n",filename);
  if(step>0)
    file = new IEEEIO(filename,IObase::Append); // open file for append or trun
  else
    file = new IEEEIO(filename,IObase::Write);
  EnzoGridInfo *grid=(levels[0])[0];
  grid->readParticleNames(names); // read all field names
  dataout.setSize(names.getSize()); // resize data holders to field size
 printf("Particle field names are\n");
  for(int i=0;i<names.getSize();i++) {

     printf("\t%s\n",names[i]) ;
    if(strstr(names[i],"index")) iindex=i;
    else if(strstr(names[i],"position")){
      if(strchr(names[i],'x')) xindex=i;
      else if(strchr(names[i],'y')) yindex=i;
      else if(strchr(names[i],'z')) zindex=i;
    }
    else if(strstr(names[i],"creation")) cindex=i;
    else if(strstr(names[i],"mass")) dindex=i;
  }

  int numgrids=0,alltotal=0;

  // cycle through each level

  for(int l=0;l<levels.getSize();l++){ // for each level
    for(int g=0;g<(levels[l]).getSize();g++){ // for each grid at that level
      numgrids++;
      grid = (levels[l])[g];
      grid->readQuantity(names[iindex],indexval);
      grid->readQuantity(names[cindex],crindex);
      // copy into an integer array? (outindex?) or float index?
      // for now, lets ignore it?

      for(int i=0;i<crindex.getSize();i++){
	if(crindex[i]>=0){               // chooses only star particles
	 dataout[iindex].append((float)(indexval[i]));  // appends starp indices
	 outindex.append(indexval[i]);  // appends starparticle indices 
	 //       printf("indexval[%d] = %d    \n",i,indexval[i]            );
	}
      }

      grid->readQuantity(names[xindex],xc); // xcoords
      grid->readQuantity(names[yindex],yc); // ycoords
      grid->readQuantity(names[zindex],zc); // zcoords
      for(int i=0;i<crindex.getSize();i++){
	if(crindex[i]>0){  // appends star particle position vectors
	  coords.append(xc[i]);
	  coords.append(yc[i]); 
	  coords.append(zc[i]);
	  leveldata.append(grid->level);
	}
      }
      
      for(int n=0;n<names.getSize();n++){
	// now we start cycling through the dataset to read each name using
	// readQuantity
	if(strstr(names[n],"index") || strstr(names[n],"position")) 
	  continue;
	grid->readQuantity(names[n],data);
	if(n==dindex){
	  fprintf(stderr,"F:[%s]  L:%2u G:%03u %04u particles",grid->filename,l,g,data.getSize());
	  alltotal+=data.getSize();
	}
	// now copy based on its index
	//(if it is a coordinate, we need to append to a single place)
	// but it has to be accumulated first

	for(int i=0;i<crindex.getSize() && i<data.getSize();i++){

	   if(crindex[i]>=0) dataout[n].append(data[i]);
	}
      }  // end of copying loop

      fprintf(stderr," : Total Stars=%u\r",
	      dataout[dindex].getSize());

    }  // end of big for loop with g
  }  // end of big honkin' for loop with l  +

  printf("\nNumber of grids=%u total particles=%u\n",numgrids,alltotal);
  density.append(dataout[dindex].getData(),dataout[dindex].getSize());
  for(int i=0;i<dataout[dindex].getSize();i++)
    (dataout[dindex])[i] = density[i] * 2.763e+11 * 2.0/(pow(2.0,3.0*leveldata[i]));

  {

    float mrange[2];

    mrange[0] = mrange[1]= (dataout[dindex])[0];
    for(int i=0;i<dataout[dindex].getSize();i++){
      float val = (dataout[dindex])[i];
      if(val<mrange[0]) mrange[0]=val;
      if(val>mrange[1]) mrange[1]=val;
    }
    printf("\nMass Range is %g:%g\n",mrange[0],mrange[1]);
  } 
  {
    float mrange[2];
    mrange[0] = mrange[1] = density[0];
    for(int i=0;i<density.getSize();i++){
      float val = density[i];
      if(val<mrange[0]) mrange[0]=val;
      if(val>mrange[1]) mrange[1]=val;
    }
    printf("\nDensity Range is %g:%g\n",mrange[0],mrange[1]);
  }
  
  /* Rescale the indices */
  {
    int minindex=outindex[0],maxindex=outindex[0];
    for(int i=0;i<outindex.getSize();i++){
      int val=outindex[i];
      if(minindex>val) minindex=val;
      if(maxindex<val) maxindex=val;
    }

    for(int i=0;i<outindex.getSize();i++){
      outindex[i]-=minindex;
      (dataout[iindex])[i]=(float)(outindex[i]);
    }
  }
  // Now we need to reindex in the Stars Data
  {
    int sz=dataout.getSize();
    dataout.setSize(dataout.getSize() + this->StarData.getSize());
    for(int i=0;i<StarData.getSize();i++,sz++){
      dataout[sz].setSize(outindex.getSize());
      char *s = new char[strlen(StarDataNames[i])+1];
      strcpy(s,StarDataNames[i]);
      names.append(s);
      mapIndexedData(outindex.getSize(),outindex.getData(),this->StarData[i].getData(),dataout[sz].getData());
    }
  }
  
  // Now that we have big-honking arrays, its time to dump them to disk
  puts("now write the stuff out!\n");
  int ldims[2];
  ldims[0]=3;
  ldims[1]=coords.getSize()/3;
  printf("\nStars Totalsize = %u particles\nWriting\n",coords.getSize()/3);
  file->write(IObase::Float32,2,ldims,coords.getData());
  for(int f=0;f<names.getSize();f++){
    if(!strstr(names[f],"position") && f!=iindex)
      file->writeAttribute(names[f],
			   IObase::Float32,dataout[f].getSize(),
			   dataout[f].getData());
    delete names[f]; // delete all of the names to clean up
  }
  file->writeAttribute("level",IObase::Float32,
		       leveldata.getSize(),leveldata.getData());
  file->writeAttribute("density",IObase::Float32,
		       density.getSize(),density.getData());   
  file->writeAttribute("findex",IObase::Float32,
		       dataout[iindex].getSize(),dataout[iindex].getData());  
  file->writeAttribute("index",IObase::Int32,
		       outindex.getSize(),outindex.getData());
  // should provide an integer index in output as well eventually
  delete file;
  return 1;
}





