#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "EnzoGridInfo.hh"
//#include "EnzoGridHierarchy.hh"

char *EnzoGridInfo::dirprefix=0;

void EnzoGridInfo::assignLevel(int lev){
  // printf("lev=%d, level=%d\n",lev,level);
  level = lev;
  if(next){
    // puts("assign next grid");
    next->assignLevel(lev);
  }
  if(nextlev){
    // puts("assign next level");
    nextlev->assignLevel(lev+1);
  }
  dx = (maxext[0]-minext[0])/(float)(dims[0]);
}
void EnzoGridInfo::print(FILE *f){
  fprintf(f,"GridID=%d level=%d\n",gridID,level);
  fprintf(f,"\tnextGridThisLevel=%d nextGridNextLevel=%d\n",
	 nextGridThisLevel,nextGridNextLevel);
  fprintf(f,"\tScaleRatio = %g\n",dx);
}

void EnzoGridInfo::parseHierarchy(FILE *f,EnzoGridInfo **root){
  // read the hierarchy
  char line[128];
  // read until we see keyword "Grid"
  while(fgets(line,127,f) && !strstr(line,"Grid ")) {}
  if(feof(f)){
    //puts("end of file");
    return;
  }
  // stop reading on blank line
  char *item = strchr(line,'=');
  item++; // go one past =
  sscanf(item,"%u",&gridID);
  while(fgets(line,127,f)!=NULL && (strlen(line)>4)){
    // we've got something to read... compare to all possibilities
    item = strchr(line,'\n');
    if(item) *item='\0'; // nuke the \n at the end of the line
    item = strchr(line,'=');
    item++;
    //printf("Grid[%u] Line=[%s] item=[%s]\n",gridID,line,item);
    if(strstr(line,"GridRank ")){
      sscanf(item,"%u",&rank);
    }
    else if(strstr(line,"GridDimension ")){
      sscanf(item,"%u %u %u",dims,dims+1,dims+2);
      dims[0]-=6; dims[1]-=6; dims[2]-=6;
    }
    else if(strstr(line,"GridLeftEdge ")){
      sscanf(item,"%lf %lf %lf",minext,minext+1,minext+2);
    }
    else if(strstr(line,"GridRightEdge ")){
      sscanf(item,"%lf %lf %lf",maxext,maxext+1,maxext+2);
    }
    else if(strstr(line,"BaryonFileName ")){
      // printf("got baryonfilename\n");
      while(*item==' ' && *item!='\0') item++;
      if(!EnzoGridInfo::dirprefix){
	filename = new char[strlen(item)+1];
	strcpy(filename,item);
      }
      else {
	filename = new char[strlen(EnzoGridInfo::dirprefix)+strlen(item)+1];
	sprintf(filename,"%s%s",EnzoGridInfo::dirprefix,item);
      }
      //printf("\t[%s]\n",filename);
    }
    else if(strstr(line,"NextGridThisLevel")){
      int srcindex,dstindex;
      // read in source and dest for this
      // how the hell to parse this!
      char *s = strchr(line,'[');
      s++;
      char *sx = strchr(s,']');
      if(sx) *sx='\0';
      //printf("\tsx=%s item=%s\n",s,item);
      srcindex = atoi(s);
      dstindex = atoi(item); 
      //printf("nextGridThisLevel[%u] srcindex=%u dstindex=%u *\n",
      //	 gridID,srcindex,dstindex);
      if(srcindex!=gridID){
	root[srcindex-1]->nextGridThisLevel = dstindex;
      }
      else{
	//printf("std write\n");
	nextGridThisLevel=dstindex;
      }
    }
    else if(strstr(line,"NextGridNextLevel")){
      int srcindex,dstindex;
      // read in source and dest for this
      // is source always < dest?
      char *s = strchr(line,'[');
      srcindex = atoi(s+1);
      dstindex = atoi(item);
      //printf("nextGridNextLevel[%u] srcindex=%u dstindex=%u *\n",
      //       gridID,srcindex,dstindex);
      if(srcindex!=gridID)
	root[srcindex-1]->nextGridNextLevel = dstindex;
      else
	nextGridNextLevel=dstindex;
    }
  } 
//puts("done");
}

int EnzoGridInfo::readFieldNames(FlexArray<char *> &fieldnames){
  HDFIO *in = new HDFIO(filename,IObase::Read);
  if(!in->isValid()) { 
    printf("could not open [%s]\n",filename); 
    delete in;
    return 0;
  } // failure
  char name[128];
  int lrank,ldims[3];
  IObase::DataType dt;
  int nitems = in->nDatasets();
  for(int i=0;i<nitems;i++){
    char *s;
    in->readInfo((char *)name,dt,lrank,ldims);
    if(lrank<3) continue; // only fields here
    s = new char[strlen(name)+1];
    strcpy(s,name);
    fieldnames.append(s);
  }
  delete in;
  return nitems;
}

int EnzoGridInfo::readParticleNames(char **fieldnames,int &nitems){
  nitems=0;
  HDFIO *in = new HDFIO(filename,IObase::Read);
  if(!in->isValid()) { delete in; return 0;} // failure
  nitems = in->nDatasets();
  char name[64];
  int lrank,ldims[3];
  IObase::DataType dt;
  for(int i=0;i<nitems;i++){
    // JMS: Fix FLEXIO!!!!!!!!!!!!!!!
    in->readInfo(name,dt,lrank,ldims);
    if((!strstr(name,"particle") && !strstr(name,"creation") && !strstr(name,"dynamical") && !strstr(name,"metallicity")) || lrank!=1) continue; // only fields here
    fieldnames[i] = new char[strlen(name)+1];
    strcpy(fieldnames[i],name);
  }
  delete in;
  return nitems;
}
int EnzoGridInfo::readParticleNames(FlexArray<char *> &fieldnames){
  HDFIO *in = new HDFIO(filename,IObase::Read);
  if(!in->isValid()) { 
    printf("could not open [%s]\n",filename); 
    delete in;
    return 0;
  } // failure
  char name[64];
  int lrank,ldims[3];
  IObase::DataType dt;
  int nitems = in->nDatasets();
  for(int i=0;i<nitems;i++){
    char *s;
    in->readInfo(name,dt,lrank,ldims);
    if((!strstr(name,"particle") && !strstr(name,"creation") && !strstr(name,"dynamical") && !strstr(name,"metallicity")) || lrank!=1) continue; // only particle fields here
    s = new char[strlen(name)+1];
    strcpy(s,name);
    fieldnames.append(s);
  }
  delete in;
  return nitems;
}

int EnzoGridInfo::readFieldNames(char **fieldnames,int &nitems){
  nitems=0;
  HDFIO *in = new HDFIO(filename,IObase::Read);
  if(!in->isValid()) { delete in; return 0;} // failure
  nitems = in->nDatasets();
  char name[64];
  int lrank,ldims[3];
  IObase::DataType dt;
  for(int i=0;i<nitems;i++){
    // JMS: Fix FLEXIO!!!!!!!!!!!!!!!
    in->readInfo(name,dt,lrank,ldims);
    if(lrank<3) continue; // only fields here
    fieldnames[i] = new char[strlen(name)+1];
    strcpy(fieldnames[i],name);
  }
  delete in;
  return nitems;
}

int EnzoGridInfo::readNames(FlexArray<char *> &fieldnames){
  HDFIO *in = new HDFIO(filename,IObase::Read);
  if(!in->isValid()) { 
    printf("could not open [%s]\n",filename); 
    delete in;
    return 0;
  } // failure
  char name[64];
  int lrank,ldims[3];
  IObase::DataType dt;
  int nitems = in->nDatasets();
  for(int i=0;i<nitems;i++){
    char *s;
    in->readInfo(name,dt,lrank,ldims);
    s = new char[strlen(name)+1];
    strcpy(s,name);
    fieldnames.append(s);
  }
  delete in;
  return nitems;
}

int EnzoGridInfo::readNames(char **fieldnames,int &nitems){
  nitems=0;
  HDFIO *in = new HDFIO(filename,IObase::Read);
  if(!in->isValid()) { delete in; return 0;} // failure
  nitems = in->nDatasets();
  char name[64];
  int lrank,ldims[3];
  IObase::DataType dt;
  for(int i=0;i<nitems;i++){
    // JMS: Fix FLEXIO!!!!!!!!!!!!!!!
    in->readInfo(name,dt,lrank,ldims);
    fieldnames[i] = new char[strlen(name)+1];
    strcpy(fieldnames[i],name);
  }
  delete in;
  return nitems;
}

void *EnzoGridInfo::readGrid(char *gridname,FlexArray<int> &data){
  HDFIO *in = new HDFIO(filename,IObase::Read);
  data.setSize(0);
  if(!in->isValid()) { delete in; return 0;}
  int ndata = in->nDatasets(); // gridfile
  int lrank,ldims[3];
  IObase::DataType dt;
  char dataname[64];
  for(int i=0;i<ndata;i++){
    // scan for dataname
    in->readInfo(dataname,dt,lrank,ldims);
    if(!strcmp(dataname,gridname)){
      // allocate space for data
      data.setSize(IObase::nElements(lrank,ldims));
      in->read(data.getData()); // read the data
      delete in;
      return data.getData();
    }
  }
  delete in;
  return 0;
}
void *EnzoGridInfo::readGrid(char *gridname,FlexArray<float> &data){
  HDFIO *in = new HDFIO(filename,IObase::Read);
  data.setSize(0);
  if(!in->isValid()){delete in;  return 0;}
  int ndata = in->nDatasets(); // gridfile
  int lrank,ldims[3];
  IObase::DataType dt;
  char dataname[64];
  for(int i=0;i<ndata;i++){
    // scan for dataname
    in->readInfo(dataname,dt,lrank,ldims);
    if(!strcmp(dataname,gridname)){
      // allocate space for data
      // float *data = new float[IObase::nElements(lrank,ldims)];
      data.setSize(IObase::nElements(rank,dims));
      in->read(data.getData()); // read the data
      delete in;
      return data.getData();
    }
  }
  delete in;
  return 0;
}

void *EnzoGridInfo::readGrid(char *gridname){
  HDFIO *in = new HDFIO(filename,IObase::Read);
  if(!in->isValid()) {delete in; return 0;}
  int ndata = in->nDatasets(); // gridfile
  int lrank,ldims[3];
  IObase::DataType dt;
  char dataname[64];
  for(int i=0;i<ndata;i++){
    // scan for dataname
    // JMS: Fix FlexIO!!!!!!!!!!!!!!!
    in->readInfo(dataname,dt,lrank,ldims);
    if(!strcmp(dataname,gridname)){
      // allocate space for data
      float *data = new float[IObase::nElements(lrank,ldims)];
      in->read(data); // read the data
      delete in;
      return data;
    }
  }
  delete in;
  return 0;
}
int EnzoGridInfo::readQuantity(char *name,
					FlexArray<int> &data){
  HDFIO *in = new HDFIO(filename,IObase::Read);
  data.setSize(0);
  if(!in->isValid()) {delete in; return 0;}
  for(int i=0;i<in->nDatasets();i++){
    char tname[64];
    int lrank,ldims[3];
    IObase::DataType datatype;
    in->readInfo(tname,datatype,lrank,ldims);
    if(lrank>1) continue; // not a 1D quantity
    if(!strcmp(tname,name)){
      // match (read quantity)
      data.setSize(ldims[0]);
      in->read(data.getData());
      delete in;
      return 1; // success
    }
  }
  delete in;
  return 0;
}

int EnzoGridInfo::readQuantity(char *name,
					FlexArray<float> &data){
  HDFIO *in = new HDFIO(filename,IObase::Read);
  data.setSize(0);
  if(!in->isValid()) { delete in; return 0;}
  for(int i=0;i<in->nDatasets();i++){
    char tname[64];
    int lrank,ldims[3];
    IObase::DataType datatype;
    in->readInfo(tname,datatype,lrank,ldims);
    if(lrank>1) continue; // failure... not a 1D quantity
    //s=tname+9;
    if(!strcmp(tname,name)){
      // match (read quantity)
      data.setSize(ldims[0]);
      in->read(data.getData());
      delete in;
      return 1;
    }
  }
  delete in;
  return 0;
}
