#ifndef __ENZOGRIDINFO_HH_
#define __ENZOGRIDINFO_HH_
#include <stdio.h>
#include <FlexArrayTmpl.H>
#include <HDFIO.hh>
#include <IEEEIO.hh>
#include <AMRwriter.hh>

struct EnzoGridInfo {
  int rank,dims[3];
  int level; // must guestimate this using parent node info recursively
  double minext[3],maxext[3],dx;
  int ngrids;
  EnzoGridInfo *nextlev,*prevlev,*next,*prev; // should have first
  // should also have immediate parent/child relationships
  // start with nchildren
  int gridID,nextGridNextLevel,nextGridThisLevel;
  FlexArray<EnzoGridInfo*> children;
  EnzoGridInfo *parent;
  char *filename;
  static char *dirprefix;
  //---Utils------------
  int isValid() { if(!filename) return 0; else return 1;}
  inline int contains(EnzoGridInfo *g);
  inline int contains(float x,float y,float z);
  void assignLevel(int lev);
  void print(FILE *f);
  //-------------------ExtractInformation----------------------
  int readParticleNames(FlexArray<char *> &fieldnames); 
  int readParticleNames(char **fieldnames,int &nitems);
  int readFieldNames(FlexArray<char *> &fieldnames); 
  int readFieldNames(char **fieldnames,int &nitems);
  int readNames(FlexArray<char *> &fieldnames); 
  int readNames(char **fieldnames,int &nitems);
  void *readGrid(char *gridname,FlexArray<int> &data);
  void *readGrid(char *gridname,FlexArray<float> &data);
  void *readGrid(char *gridname);
  // Read Quantity is redundant (same as readGrid!)
  // actuallly it is different... it can only read 1D datasets.
  // this has to be eliminated....
  int readQuantity(char *name,FlexArray<float> &data); // how does this differ from readGrid?
  int readQuantity(char *name,FlexArray<int> &data); // read integer data
  void parseHierarchy(FILE *f,EnzoGridInfo **root);

  //-------------------Constructor-Destructor------------------
  EnzoGridInfo(FILE *hierarchyfile,EnzoGridInfo **root):filename(0),
    next(0),nextlev(0),level(-1),
    prev(0),prevlev(0),
    gridID(-1),nextGridNextLevel(-1),nextGridThisLevel(-1){
    parseHierarchy(hierarchyfile,root);
  }
  ~EnzoGridInfo(){
  }
};

int EnzoGridInfo::contains(EnzoGridInfo *g){
  if(minext[0]<g->minext[0] && 
     minext[1]<g->minext[1] &&
     minext[2]<g->minext[2] &&
     maxext[0]<g->maxext[0] &&
     maxext[1]<g->maxext[1] &&
     maxext[2]<g->maxext[2]) return 1;
  else return 0;
}

int EnzoGridInfo::contains(float x,float y,float z){
  if(minext[0]<x && 
     minext[1]<y &&
     minext[2]<z &&
     maxext[0]<x &&
     maxext[1]<y &&
     maxext[2]<z) return 1;
  else return 0;
}
#endif // EnzoGridInfo

