#include <stdio.h>
#include <stdlib.h>
#include <FlexArrayTmpl.H>
#include <HDFIO.hh>
#include <IEEEIO.hh>
#include "EnzoGridHierarchy.hh"
#include "EnzoGridInfo.hh"

struct CmdlnOptions {
  char *programname;
  char *filename,*outfileprefix;
  char *path;
  typedef char *charP;
  int outputBboxASCII;
  int outputInfo,outputGasParticles,outputStarParticles,outputDMParticles,outputMPI,outputJad; // info-only
  int genParticles;
  // int nFieldNames;
  char fieldname[64];
  int append,truncate; // append mode vs. truncate mode for writes
  // select specific levels  first:last  or first,second,third or ~first
  void setDefaults(){
    programname="";
    filename=0;
    outfileprefix=0;
    path = 0;
    genParticles=0;
    outputJad=0;
    outputMPI=0;
    outputBboxASCII=0;
    outputGasParticles=outputDMParticles=outputStarParticles=0;
    outputInfo=0;
    fieldname[0]='\0';
  }
  CmdlnOptions(){
    setDefaults();
  }
  CmdlnOptions(int argc,char *argv[]){
    setDefaults();
    parse(argc,argv);
  }
  ~CmdlnOptions(){
    if(path) delete path;
  }
  void parse(int argc,char *argv[]){
    programname=argv[0];
    // printf("parse file\n");
    for(int i=1;i<argc;i++){
      char *str=argv[i];
      if(*str=='-'){
	char *val=argv[i+1];
	str++;
	if(!strcmp(str,"gas") || *str=='g') {
	  // write only particles
	  outputGasParticles=1;
	}
	else if(!strcmp(str,"dm") || *str=='d') {
	  // write only particles
	  outputDMParticles=1;
	}
	else if(!strncmp(str,"star",5) || *str=='s') {
	  // write only particles
	  outputStarParticles=1;
	}
	else if(!strcmp(str,"jad") || *str=='j'){
	  outputJad=1;
	}
	else if(!strcmp(str,"mpi") || *str=='m'){
	  outputMPI=1;
	}

	else if(!strcmp(str,"info") || *str=='i'){
	  outputInfo=1;
	}
	else if(!strcmp(str,"out") || *str=='o'){
	  outfileprefix = val;
	  i++;
	}
	else if(!strcmp(str,"truncate") || *str=='t') {
	  truncate = 1; append=0;
	}
	else if(!strcmp(str,"append") || *str=='a'){
	  truncate = 0; append=1;
	}
	else if(!strcmp(str,"field") || *str=='f'){
	  i++;
	  strcpy(fieldname,val);
	}
      }
      else{ // need to be able to append files to a list
	// its the infilename which should be a .hierarchy file
	filename = str; // just point directly (why copy?)
	// got to get more sophisticated here
	if(strchr(filename,'/')) {
	  path = new char[strlen(filename)+1];
	  strcpy(path,filename);
	  char *s = strrchr(path,'/');
	  s[1] = '\0';
	  printf("path=[%s] filename=[%s]\n",path,filename);
	}
	else {
	  path = new char[1];
	  *path = '\0';
	  printf("path=[%s] filename=[%s]\n",path,filename);
	}
      }
    }
  }
};

int main(int argc,char *argv[]){
  CmdlnOptions cmdln(argc,argv);
  EnzoGridHierarchy hierarchy;

  if(cmdln.path) EnzoGridInfo::dirprefix = cmdln.path;
  puts("-------------Read Hierarchy----------------");
  if(!hierarchy.readHierarchy(cmdln.filename)){
    puts("exiting");
    exit(0);
  }
  if(cmdln.outputInfo){
    puts("-------------Print Info------------");
    hierarchy.outputInfo();
  }
  if(cmdln.outputGasParticles){
    char outfilename[128],tempoutfilename[128];
    if(cmdln.outfileprefix)
      sprintf(tempoutfilename,"%sgas.",cmdln.outfileprefix);
    else {
      if(cmdln.path)
	strcpy(tempoutfilename,cmdln.filename+strlen(cmdln.path));
      else
	strcpy(tempoutfilename,cmdln.filename);
      char *s = strrchr(tempoutfilename,'.'); // strip off end of name
      *s='\0';
    }
    char *fieldname,*prefix;
    if(cmdln.fieldname && *cmdln.fieldname!='\0') fieldname = cmdln.fieldname;
    else fieldname = "all";
    prefix = tempoutfilename; 
    sprintf(outfilename,"%s%s.gas.raw",prefix,fieldname);
    printf("Output filename=[%s]\n",outfilename);
    if(!strcmp(fieldname,"all") || !strcmp(fieldname,"All"))
	hierarchy.writeAllGasParticles(outfilename, cmdln.append?1:0);
    else
      hierarchy.writeGasParticles(outfilename,fieldname); // stop zero in each case
  }
  if(cmdln.outputJad){
    char outfilename[128];
    if(cmdln.outfileprefix)
      sprintf(outfilename,"%samr",cmdln.outfileprefix);
    else {
      if(cmdln.path)
	strcpy(outfilename,cmdln.filename+strlen(cmdln.path));
      else
	strcpy(outfilename,cmdln.filename);
      char *s = strrchr(outfilename,'.'); // strip off end of name
      *s='\0';
    }
    printf("Output filename=[%s]\n",outfilename);
    //exit(0);
    if(*cmdln.fieldname){
      if(!strcmp(cmdln.fieldname,"all") || !strcmp(cmdln.fieldname,"All"))
	hierarchy.writeAllJad(outfilename, cmdln.append?1:0);
      else
	hierarchy.writeJad(outfilename,cmdln.fieldname); // stop zero in each case
    }
    else
      hierarchy.writeJad(outfilename,"Density"); // stop zero in each case
    //hierarchy.writeAllJad(outfilename); // stop zero in each case
  } 

  if(!cmdln.outputMPI){
  if(cmdln.outputDMParticles){
    char outfilename[128],tempoutfilename[128];
    if(cmdln.outfileprefix)
      sprintf(tempoutfilename,"%sdm.",cmdln.outfileprefix);
    else {
      if(cmdln.path)
	strcpy(tempoutfilename,cmdln.filename+strlen(cmdln.path));
      else
	strcpy(tempoutfilename,cmdln.filename);
      char *s = strrchr(tempoutfilename,'.'); // strip off end of name
      *s='\0';
    }
    char *fieldname,*prefix;
    if(cmdln.fieldname && *cmdln.fieldname!='\0') fieldname = cmdln.fieldname;
    else fieldname = "all";
    //if(cmdln.outfileprefix) prefix = cmdln.outfileprefix;
    //else prefix = tempoutfilename;
    prefix = tempoutfilename;
    sprintf(outfilename,"%s%s.dm.raw",prefix,fieldname);
    printf("Output filename=[%s]\n",outfilename);
    if(!strcmp(fieldname,"all") || !strcmp(fieldname,"All")){
      puts("write it all out");
      hierarchy.writeAllDMParticles(outfilename, cmdln.append?1:0);
      puts("done writing");
    }
    else {
      printf("write only the field %s\n",fieldname);
      hierarchy.writeDMParticles(outfilename,fieldname); // stop zero in each case
    }
  } 
  if(cmdln.outputStarParticles){
    char outfilename[128],tempoutfilename[128];
    if(cmdln.outfileprefix)
      sprintf(tempoutfilename,"%sstar.",cmdln.outfileprefix);
    else {
      if(cmdln.path)
	strcpy(tempoutfilename,cmdln.filename+strlen(cmdln.path));
      else
	strcpy(tempoutfilename,cmdln.filename);
      char *s = strrchr(tempoutfilename,'.'); // strip off end of name
      *s='\0';
    }
    char *fieldname,*prefix;
    if(cmdln.fieldname && *cmdln.fieldname!='\0') fieldname = cmdln.fieldname;
    else fieldname = "all";
    if(cmdln.outfileprefix) prefix = cmdln.outfileprefix;
    else prefix = tempoutfilename;
    sprintf(outfilename,"%s%s.star.raw",prefix,fieldname);
    printf("Output filename=[%s]\n",outfilename);
    if(!strcmp(fieldname,"all") || !strcmp(fieldname,"All"))
	hierarchy.writeAllStarParticles(outfilename, cmdln.append?1:0);
    else
      hierarchy.writeStarParticles(outfilename,fieldname); // stop zero in each case
  }
  }

  if(cmdln.outputMPI){  // for mpi version
  if(cmdln.outputDMParticles){
    char outfilename[128],tempoutfilename[128];
    if(cmdln.outfileprefix)
      sprintf(tempoutfilename,"%sdm.",cmdln.outfileprefix);
    else {
      if(cmdln.path)
	strcpy(tempoutfilename,cmdln.filename+strlen(cmdln.path));
      else
	strcpy(tempoutfilename,cmdln.filename);
      char *s = strrchr(tempoutfilename,'.'); // strip off end of name
      *s='\0';
    }
    char *fieldname,*prefix;
    if(cmdln.fieldname && *cmdln.fieldname!='\0') fieldname = cmdln.fieldname;
    else fieldname = "all";
    //if(cmdln.outfileprefix) prefix = cmdln.outfileprefix;
    //else prefix = tempoutfilename;
    prefix = tempoutfilename;
    sprintf(outfilename,"%s%s.dm.raw",prefix,fieldname);
    printf("Output filename=[%s]\n",outfilename);
    if(!strcmp(fieldname,"all") || !strcmp(fieldname,"All")){
      puts("write it all out");
      hierarchy.writeAllDMParticlesMPI(outfilename, cmdln.append?1:0);
      puts("done writing");
    }
    else {
      printf("write only the field %s\n",fieldname);
      hierarchy.writeDMParticles(outfilename,fieldname); // stop zero in each case
    }
  } 
  if(cmdln.outputStarParticles){
    char outfilename[128],tempoutfilename[128];
    if(cmdln.outfileprefix)
      sprintf(tempoutfilename,"%sstar.",cmdln.outfileprefix);
    else {
      if(cmdln.path)
	strcpy(tempoutfilename,cmdln.filename+strlen(cmdln.path));
      else
	strcpy(tempoutfilename,cmdln.filename);
      char *s = strrchr(tempoutfilename,'.'); // strip off end of name
      *s='\0';
    }
    char *fieldname,*prefix;
    if(cmdln.fieldname && *cmdln.fieldname!='\0') fieldname = cmdln.fieldname;
    else fieldname = "all";
    if(cmdln.outfileprefix) prefix = cmdln.outfileprefix;
    else prefix = tempoutfilename;
    sprintf(outfilename,"%s%s.star.raw",prefix,fieldname);
    printf("Output filename=[%s]\n",outfilename);
    if(!strcmp(fieldname,"all") || !strcmp(fieldname,"All"))
	hierarchy.writeAllStarParticlesMPI(outfilename, cmdln.append?1:0);
    else
      hierarchy.writeStarParticles(outfilename,fieldname); // stop zero in each case
  }
  }








}

/*
  Get grid info (the names of grids)
     Grid: Gridname
     Particles: Particlename (actually component names for particles)
  Output as JAD
  Output as ParticlesHDF (nonoverlapping)
  Output as ParticlesUCD (nonoverlapping)
  Output as UCD (hexahedral cells)
 */
